/* nm-novellvpn-propertiesui.c : GNOME UI dialogs for configuring NovellVPN connections
 *
 * Copyright (C) 2006 Sureshkumar T <tsureshkumar@novell.com>
 * Based on work by Tim Niemueller <tim@niemueller.de>
 *                  David Zeuthen, <davidz@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n-lib.h>
#include <string.h>
#include <stdlib.h>

#include "nm-novellvpn-propertiesui.h"

#define NM_VPN_API_SUBJECT_TO_CHANGE

#include <NetworkManager/nm-vpn-ui-interface.h>

#define KEY_GWTYPE     "gateway-type"
#define KEY_CONN_NAME  "connection_name"
#define KEY_GATEWAY    "remote"
#define KEY_USER_NAME  "username"
#define KEY_GROUP_NAME "group-name"
#define KEY_DHGROUP    "dhgroup"

#define NM_NOVELLVPN_GWID_NORTEL          0
#define NM_NOVELLVPN_GWTYPE_NORTEL        "nortel"

typedef enum {
  dhgroup_0,
  DH1,
  DH2
} DHGroup;


static void
load_widgets_and_handlers (NovellVPNPropertiesUI *self, const gchar *glade_file ) 
{
  g_assert (self->editable_changed != NULL);

  self->xml = glade_xml_new (glade_file, NULL, GETTEXT_PACKAGE);

  if (self->xml != NULL) {
    self->widget = glade_xml_get_widget(self->xml, "nm-novellvpn-widget");
    self->w_connection_name   = GTK_ENTRY (glade_xml_get_widget (self->xml, "nm-novellvpn-connection-name"));
    self->w_gatewaytype       = GTK_COMBO_BOX (glade_xml_get_widget (self->xml, "nm-novellvpn-gatewaytype"));
    self->w_gateway           = GTK_ENTRY (glade_xml_get_widget (self->xml, "nm-novellvpn-gateway"));
    self->w_username          = GTK_ENTRY (glade_xml_get_widget (self->xml, "nm-novellvpn-username"));
    self->w_groupname         = GTK_ENTRY (glade_xml_get_widget (self->xml, "nm-novellvpn-groupname"));
    self->w_dhgroup           = GTK_COMBO_BOX (glade_xml_get_widget (self->xml, "nm-novellvpn-dhgroup"));

    self->w_opt_info_expander      = GTK_EXPANDER (glade_xml_get_widget (self->xml, 
									 "novellvpn-optional-information-expander"));

    gtk_signal_connect (GTK_OBJECT (self->w_connection_name), 
			"changed", GTK_SIGNAL_FUNC (*(self->editable_changed)), self->editable_changed_user_data);
    gtk_signal_connect (GTK_OBJECT (self->w_gateway), 
			"changed", GTK_SIGNAL_FUNC (self->editable_changed), self->editable_changed_user_data);
    gtk_signal_connect (GTK_OBJECT (self->w_username), 
			"changed", GTK_SIGNAL_FUNC (self->editable_changed), self->editable_changed_user_data);
    gtk_signal_connect (GTK_OBJECT (self->w_groupname), 
			"changed", GTK_SIGNAL_FUNC (self->editable_changed), self->editable_changed_user_data);

    /* make the widget reusable */
    gtk_signal_connect (GTK_OBJECT (self->widget), "delete-event", 
			GTK_SIGNAL_FUNC (gtk_widget_hide_on_delete), NULL);

    g_assert (self->widget != NULL );
    g_assert (self->w_connection_name != NULL );
    g_assert (self->w_gatewaytype != NULL );
    g_assert (self->w_gateway != NULL );
    g_assert (self->w_username != NULL );
    g_assert (self->w_groupname != NULL );
    g_assert (self->w_dhgroup != NULL );
    g_assert (self->w_opt_info_expander != NULL);
  }
}

void 
novellvpnui_clear_widget (NovellVPNPropertiesUI *self)
{
  gtk_combo_box_set_active (GTK_COMBO_BOX (self->w_gatewaytype), 0);

  gtk_entry_set_text (self->w_connection_name, "");
  gtk_entry_set_text (self->w_gateway,  "");
  gtk_entry_set_text (self->w_username,  "");
  gtk_entry_set_text (self->w_groupname,  "");
  
  gtk_combo_box_set_active (GTK_COMBO_BOX (self->w_dhgroup), 0);

  gtk_expander_set_expanded (self->w_opt_info_expander, FALSE);
}

gchar *
novellvpnui_get_connection_name (NovellVPNPropertiesUI *self)
{
  const char *name =  gtk_entry_get_text (self->w_connection_name);
  if (name != NULL)
    return g_strdup (name);
  else
    return NULL;
}

void
novellvpnui_set_connection_name (NovellVPNPropertiesUI *self, const gchar *name)
{
  if (name != NULL)
    gtk_entry_set_text (self->w_connection_name, name);
  else
    gtk_entry_set_text (self->w_connection_name, "");
}

GSList *
novellvpnui_get_properties (NovellVPNPropertiesUI *self)
{
  GSList *data = NULL;
  const char *gwtype;
  const char *gw;
  const char *username;
  const char *groupname;
  DHGroup dhgroup;


  gw                     = gtk_entry_get_text (self->w_gateway);
  username               = gtk_entry_get_text (self->w_username);
  groupname              = gtk_entry_get_text (self->w_groupname);
  
  switch ( gtk_combo_box_get_active (GTK_COMBO_BOX (self->w_gatewaytype)) ) {
  case NM_NOVELLVPN_GWID_NORTEL:
  default:
    gwtype = NM_NOVELLVPN_GWTYPE_NORTEL;
  }

  switch ( gtk_combo_box_get_active (GTK_COMBO_BOX (self->w_dhgroup))) {
  case 1:
    dhgroup = DH2;
    break;
  case 0:
  default:
    dhgroup = DH1;
  }

  data = NULL;
  data = g_slist_append (data, g_strdup (KEY_GWTYPE));
  data = g_slist_append (data, g_strdup (gwtype));
  data = g_slist_append (data, g_strdup (KEY_GATEWAY));
  data = g_slist_append (data, g_strdup (gw));
  data = g_slist_append (data, g_strdup (KEY_USER_NAME));
  data = g_slist_append (data, g_strdup (username));
  data = g_slist_append (data, g_strdup (KEY_GROUP_NAME));
  data = g_slist_append (data, g_strdup (groupname));
  data = g_slist_append (data, g_strdup (KEY_DHGROUP));
  data = g_slist_append (data, dhgroup == DH2 ? g_strdup ("dh2") : g_strdup("dh1"));
  return data;

}

void
novellvpnui_set_properties (NovellVPNPropertiesUI *self, GSList *properties)
{
  GSList *i;
  gboolean should_expand = FALSE;

  for (i = properties; i != NULL && g_slist_next (i) != NULL; i = g_slist_next (g_slist_next (i))) {
    const char *key;
    const char *value;

    key = i->data;
    value = (g_slist_next (i))->data;

    if (strcmp (key, KEY_GWTYPE) == 0) {
      gint type_cbox_sel = 0;

      if ( strcmp (value, KEY_GWTYPE) == 0 ) {
	type_cbox_sel = NM_NOVELLVPN_GWID_NORTEL;
      } 
      gtk_combo_box_set_active (GTK_COMBO_BOX (self->w_gatewaytype), type_cbox_sel);
    } else if (strcmp (key, KEY_GATEWAY) == 0) {
      gtk_entry_set_text (self->w_gateway, value);
    } else if (strcmp (key, KEY_USER_NAME) == 0) {
      gtk_entry_set_text (self->w_username, value);
    } else if (strcmp (key, KEY_GROUP_NAME) == 0) {
      gtk_entry_set_text (self->w_groupname, value);
      should_expand = TRUE;
    } else if (strcmp (key, KEY_DHGROUP) == 0) {
      DHGroup dh_group = strcmp (value, "dh2") == 0 ? DH2 : DH1;
      gtk_combo_box_set_active (GTK_COMBO_BOX (self->w_gatewaytype), dh_group - 1);
      should_expand = TRUE;
    }
  }

  gtk_expander_set_expanded (self->w_opt_info_expander, should_expand);
  gtk_container_resize_children (GTK_CONTAINER (self->widget));
}

gboolean
novellvpnui_validate (NovellVPNPropertiesUI *self)
{
  gboolean is_valid;
  const char *connectionname;
  const char *remote;
  const char *groupname;

  connectionname         = gtk_entry_get_text (self->w_connection_name);
  remote                 = gtk_entry_get_text (self->w_gateway);

  is_valid = FALSE;

  if ( (strlen (connectionname) == 0) ||
       (strlen (remote) == 0) ||
       (strstr (remote, " ") != NULL)  ||
       (strstr (remote, "\t") != NULL) ) {

    is_valid = FALSE;

  } else {
    const char *username;

    username    = gtk_entry_get_text (self->w_username);
    groupname              = gtk_entry_get_text (self->w_groupname);

    if (strlen (groupname) > 0 &&
	(strlen (username) > 0))
      is_valid = TRUE;
  } 

  return is_valid;
}

gboolean
novellvpnui_init (NovellVPNPropertiesUI *self) 
{
  gchar *glade_file = g_strdup_printf ("%s/%s", GLADEDIR, "nm-novellvpn-dialog.glade");
  load_widgets_and_handlers (self, (const gchar *) glade_file);
  g_free( glade_file );

  if (self->xml == NULL)
    return FALSE;

  novellvpnui_clear_widget (self);
  return TRUE;
}
