/* nm-novellvpn-vpnui-impl.c : Factory Implementations for configuring 
 *                             NovellVPN connections
 *
 * Copyright (C) 2006 Sureshkumar T <tsureshkumar@novell.com>
 * Based on work by Tim Niemueller <tim@niemueller.de>
 *                  David Zeuthen, <davidz@redhat.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <glib/gi18n-lib.h>
#include <string.h>
#include <stdlib.h>

#define NM_VPN_API_SUBJECT_TO_CHANGE

#include <NetworkManager/nm-vpn-ui-interface.h>

#include "nm-novellvpn-propertiesui.h"

typedef struct _NetworkManagerVpnUIImpl NetworkManagerVpnUIImpl;

struct _NetworkManagerVpnUIImpl {
  NetworkManagerVpnUI parent;
  
  NovellVPNPropertiesUI *novell_vpn_ui;

  NetworkManagerVpnUIDialogValidityCallback callback;
  gpointer callback_user_data;
};

static gboolean
impl_is_valid (NetworkManagerVpnUI *self)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;
  return novellvpnui_validate (impl->novell_vpn_ui);
}


static void 
editable_changed (GtkEditable *editable, gpointer user_data)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) user_data;

  if (impl->callback != NULL) {
    gboolean is_valid;

    is_valid = impl_is_valid (&(impl->parent));
    impl->callback (&(impl->parent), is_valid, impl->callback_user_data);
  }
}

static const char *
impl_get_display_name (NetworkManagerVpnUI *self)
{
  return _("NovellVPN Client");
}

static const char *
impl_get_service_name (NetworkManagerVpnUI *self)
{
  return "org.freedesktop.NetworkManager.novellvpn";
}

static GtkWidget *
impl_get_widget (NetworkManagerVpnUI *self, GSList *properties, GSList *routes, const char *connection_name)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;
  NovellVPNPropertiesUI *novell_vpn_ui = impl->novell_vpn_ui;

  novellvpnui_clear_widget (novell_vpn_ui);
  novellvpnui_set_connection_name (novell_vpn_ui, connection_name);
  novellvpnui_set_properties (novell_vpn_ui, properties);

  return novell_vpn_ui->widget;
}

static char *
impl_get_connection_name (NetworkManagerVpnUI *self)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;
  return novellvpnui_get_connection_name (impl->novell_vpn_ui);
}

static GSList *
impl_get_properties (NetworkManagerVpnUI *self)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;
  return novellvpnui_get_properties (impl->novell_vpn_ui);
}

static void 
impl_set_validity_changed_callback (NetworkManagerVpnUI *self, 
				    NetworkManagerVpnUIDialogValidityCallback callback,
				    gpointer user_data)
{
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;

  impl->callback = callback;
  impl->callback_user_data = user_data;
}

static void
impl_get_confirmation_details (NetworkManagerVpnUI *self, gchar **retval)
{
  GString *buf;
  NetworkManagerVpnUIImpl *impl = (NetworkManagerVpnUIImpl *) self->data;
  NovellVPNPropertiesUI *novell_vpn_ui = impl->novell_vpn_ui;

  // This is risky, should be variable length depending on actual data!
  buf = g_string_sized_new (512);

  g_string_append (buf, _("The following NovellVPN connection will be created:"));
  g_string_append (buf, "\n\n\t");
  g_string_append_printf (buf, _("Name:  %s"), novellvpnui_get_connection_name (novell_vpn_ui));
  g_string_append (buf, "\n\n\t");
  g_string_append (buf, "\n\n");
  g_string_append (buf, _("The connection details can be changed using the \"Edit\" button."));
  g_string_append (buf, "\n");

  *retval = g_string_free (buf, FALSE);
}

GSList *
impl_get_routes (NetworkManagerVpnUI *self)
{
  return NULL; // does not use routes
}

static gboolean 
impl_import_file (NetworkManagerVpnUI *self, const char *path)
{
  return TRUE;
}

static gboolean 
impl_can_export (NetworkManagerVpnUI *self)
{
  return FALSE;
}

static gboolean 
impl_export (NetworkManagerVpnUI *self, GSList *properties, GSList *routes, const char *connection_name)
{
  return TRUE;
}

static NetworkManagerVpnUI* 
impl_create_object (void)
{
  char *glade_file;
  NetworkManagerVpnUIImpl *impl;

  impl = g_new0 (NetworkManagerVpnUIImpl, 1);
  impl->novell_vpn_ui = g_new0 (NovellVPNPropertiesUI, 1);
  impl->novell_vpn_ui->editable_changed = editable_changed;
  impl->novell_vpn_ui->editable_changed_user_data = impl;
  
  if (! novellvpnui_init (impl->novell_vpn_ui)) {
    g_free (impl->novell_vpn_ui);
    g_free (impl);
    return NULL;
  }

  impl->callback                             = NULL;

  // override methods
  impl->parent.get_display_name              = impl_get_display_name;
  impl->parent.get_service_name              = impl_get_service_name;
  impl->parent.get_widget                    = impl_get_widget;
  impl->parent.get_connection_name           = impl_get_connection_name;
  impl->parent.get_properties                = impl_get_properties;
  impl->parent.set_validity_changed_callback = impl_set_validity_changed_callback;
  impl->parent.is_valid                      = impl_is_valid;
  impl->parent.get_confirmation_details      = impl_get_confirmation_details;
  impl->parent.data                          = impl;

  impl->parent.get_routes                    = impl_get_routes;
  impl->parent.can_export                    = impl_can_export;
  impl->parent.import_file                   = impl_import_file;
  impl->parent.export                        = impl_export;
    
  return &(impl->parent);
}


NetworkManagerVpnUI* 
nm_vpn_properties_factory (void)
{
  return impl_create_object();
}
